@test_throws MethodError shuffleobs(kfolds(X1))
@test_throws MethodError splitobs(kfolds(X1))
@test_throws MethodError oversample(kfolds(X1))
@test_throws MethodError undersample(kfolds(X1))
@test_throws MethodError stratifiedobs(kfolds(X1))

@testset "kfolds with Int" begin
    @test typeof(kfolds) <: Function
    @test typeof(kfolds(10)) <: Tuple

    @test_throws ArgumentError kfolds(10,1)
    @test_throws ArgumentError kfolds(10,11)
    @test_throws ArgumentError kfolds(1)
    @test_throws ArgumentError kfolds(2)
    @test_throws ArgumentError kfolds(-1)

    for n in (5,10,15,20,50), k in (2,5)
        for (f1,f2) in (@inferred(kfolds(n,k)), @inferred(kfolds(n)))
            @test typeof(f1) <: Vector{Vector{Int}}
            @test typeof(f2) <: Vector{UnitRange{Int}}
            @test length(unique(vcat(f2...))) == n
            for (i1,i2) in zip(f1,f2)
                @test length(unique(union(i2,i1))) == n
            end
        end
    end
end

@testset "leaveout with Int" begin
    @test typeof(leaveout) <: Function
    @test typeof(leaveout(10)) <: Tuple

    @test_throws ArgumentError leaveout(10,10)
    @test_throws ArgumentError leaveout(10,6)
    @test_throws ArgumentError leaveout(10,-1)
    @test_throws ArgumentError leaveout(1)
    @test_throws ArgumentError leaveout(-1)

    for n in (5,10,15,20,50), s in (1,2)
        for (f1,f2) in (@inferred(leaveout(n,s)), @inferred(leaveout(n)))
            @test typeof(f1) <: Vector{Vector{Int}}
            @test typeof(f2) <: Vector{UnitRange{Int}}
            @test length(unique(vcat(f2...))) == n
            for (i1,i2) in zip(f1,f2)
                @test length(unique(union(i2,i1))) == n
            end
        end
    end
end

println("<HEARTBEAT>")

@testset "FoldsView constructor" begin
    @test FoldsView <: AbstractVector
    @test FoldsView <: LearnBase.DataView
    @test FoldsView{Tuple} <: LearnBase.DataView{Tuple}

    @test_reference "references/FoldsView.txt" @io2str(show(::IO, MIME"text/plain"(), kfolds(rand(10),k=5))) by=matrix_compat_isequal

    @testset "Illegal arguments" begin
        # fold indices out of bounds for the given data
        tx = rand(2,15)
        ty = rand(15)
        @test_throws DimensionMismatch FoldsView(tx, kfolds(16)...)
        @test_throws DimensionMismatch FoldsView((tx,rand(14)), kfolds(15)...)
        # number of folds must match for train and test
        f1,tmp = kfolds(15,3)
        tmp,f2 = kfolds(15,5)
        @test_throws DimensionMismatch FoldsView(tx, f1, f2)
        @test_throws DimensionMismatch FoldsView((tx,ty), f1, f2)
        # Only accept Int arrays as indices
        f1,f2 = kfolds(15)
        @test_throws ArgumentError FoldsView(tx, map(x->Float64.(x),f1), f2)
        @test_throws ArgumentError FoldsView(tx, f1, map(x->Float64.(x),f2))
        # only one fold should fail (or should it not?)
        @test_throws ArgumentError FoldsView(tx, map(x->x[1:1], kfolds(15,2))...)
    end

    # FIXME: in 0.6 change back to (Xs,ys,vars...,tuples...)
    for var in (Xs, ys, X, Xv, yv, XX, tuples[1])
        f1, f2 = kfolds(nobs(var), 5)
        for fv in (@inferred(FoldsView(var, f1, f2)),
                   @inferred(FoldsView(var, f1, f2, ObsDim.Last())),
                   FoldsView(var, f1, f2, obsdim=:last))
            @test sum(length.(fv.val_indices)) == nobs(var)
            @test fv.data === var
            @test length(fv.train_indices) === 5
            @test length(fv.val_indices)  === 5
            @test fv.train_indices === f1
            @test fv.val_indices  === f2
            for ((tr, te), i1, i2) in zip(fv, f1, f2)
                @test tr == datasubset(var, i1)
                @test te == datasubset(var, i2)
            end
        end

        f1, f2 = kfolds(nobs(var), 10)
        fv = @inferred(FoldsView(var, f1, f2))
        @test fv.data === var
        @test length(fv.train_indices) === 10
        @test length(fv.val_indices)  === 10
        @test fv.train_indices === f1
        @test fv.val_indices  === f2
        for ((tr, te), i1, i2) in zip(fv, f1, f2)
            @test tr == datasubset(var, i1)
            @test te == datasubset(var, i2)
        end

        f1, f2 = kfolds(nobs(var), 20)
        fv = @inferred(FoldsView(var, f1, f2))
        cumobs = length(fv.val_indices[1])
        for i = 2:length(fv.val_indices)
            cumobs += length(fv.val_indices[i])
            @test 7 <= length(fv.val_indices[i]) <= length(fv.val_indices[1]) <= 8
            @test 1 <= minimum(fv.val_indices[i-1]) < minimum(fv.val_indices[i]) < nobs(var)
        end
        @test cumobs == nobs(var)

        println("<HEARTBEAT>")
    end

    # check that the stored obsdim is correct
    f1, f2 = kfolds(nobs(X), 5)
    # FIXME: in 0.6 change back to (Xs,ys,vars...)
    for var in (Xs, ys, Xv, yv)
        fv = FoldsView(var, f1, f2)
        @test fv.obsdim == ObsDim.Last()
    end
    # FIXME: in 0.6 change back to (tuples...)
    for var in (tuples[1], tuples[2])
        fv = FoldsView(var, f1, f2)
        @test typeof(fv.obsdim) <: Tuple
        @test all(map(x->typeof(x)<:ObsDim.Last, fv.obsdim))
    end
end

println("<HEARTBEAT>")

@testset "FoldsView getindex, endof, length" begin
    f1, f2 = kfolds(nobs(X), 10)
    for var in vars
        fv = FoldsView(var, f1, f2)
        @test length(fv) == 10
        @test fv[end] == fv[length(fv)]
    end
    for var in (Xs, ys, vars...)
        fv = FoldsView(var, f1, f2)
        @test length(fv) == 10
        @test getobs.(fv[end]) == getobs.(fv[length(fv)])
        println("<HEARTBEAT>")
    end
    f1, f2 = kfolds(nobs(X))
    fv = FoldsView(X, f1, f2)
    @test typeof(@inferred(fv[1])) <: Tuple
    @test typeof(@inferred(fv[2:3])) <: FoldsView
    @test @inferred(getobs(fv,1)) == fv[1]
    @test @inferred(getobs(fv[2:3])) == @inferred(getindex(getobs(fv), 2:3))
    @test @inferred(getobs(fv[2:3])) == [fv[2], fv[3]]
    @test typeof(fv[1][1]) <: SubArray
    @test typeof(fv[1][2]) <: SubArray
    @test size(fv[1][1]) == (4,120)
    @test size(fv[1][2]) == (4,30)
    fv = FoldsView(X', f1, f2, obsdim=1)
    @test typeof(@inferred(fv[1])) <: Tuple
    @test typeof(fv[1][1]) <: SubArray
    @test typeof(fv[1][2]) <: SubArray
    @test size(fv[1][1]) == (120,4)
    @test size(fv[1][2]) == (30,4)
    fv = FoldsView((X',X), f1, f2, obsdim=(1,2))
    @test typeof(@inferred(fv[1])) <: Tuple
    @test typeof(fv[1][1]) <: Tuple
    @test typeof(fv[1][2]) <: Tuple
    @test typeof(fv[1][1][1]) <: SubArray
    @test typeof(fv[1][1][2]) <: SubArray
    @test typeof(fv[1][2][1]) <: SubArray
    @test typeof(fv[1][2][2]) <: SubArray
    @test size(fv[1][1][1]) == (120,4)
    @test size(fv[1][1][2]) == (4,120)
    @test size(fv[1][2][1]) == (30,4)
    @test size(fv[1][2][2]) == (4,30)
end

println("<HEARTBEAT>")

@testset "FoldsView iteration" begin
    # FIXME: in 0.6 change back to (X, Xv, yv, XX, XXX, y)
    for var in (X, Xv, yv, XX, y)
        all_train_indices = Array{Int,1}()
        all_test_indices = Array{Int,1}()
        for (train, test) in FoldsView(var, kfolds(nobs(X),10)...)
            @test nobs(train) == 135
            @test nobs(test)  == 15

            @test typeof(train) <: SubArray
            @test typeof(test) <: SubArray

            @test length(setdiff(train.indices[ndims(train)], test.indices[ndims(test)])) == 135
            @test length(setdiff(test.indices[ndims(test)], train.indices[ndims(train)])) == 15
            append!(all_train_indices, train.indices[ndims(train)])
            append!(all_test_indices, test.indices[ndims(test)])
        end
        @test length(unique(all_train_indices)) == 150
        @test length(unique(all_test_indices)) == 150
        @test length(all_test_indices) == 150
        println("<HEARTBEAT>")
    end
    for var in (Xs, ys)
        all_train_indices = Array{Int,1}()
        all_test_indices = Array{Int,1}()
        for (train, test) in FoldsView(var, kfolds(nobs(X),10)...)
            @test nobs(train) == 135
            @test nobs(test)  == 15

            @test typeof(train) <: DataSubset
            @test typeof(test) <: DataSubset

            @test length(setdiff(train.indices, test.indices)) == 135
            @test length(setdiff(test.indices, train.indices)) == 15

            append!(all_train_indices, train.indices)
            append!(all_test_indices, test.indices)
        end
        @test length(unique(all_train_indices)) == 150
        @test length(unique(all_test_indices)) == 150
        @test length(all_test_indices) == 150
    end
    for var in (Xs, ys)
        for (train, test) in FoldsView(var, kfolds(nobs(X),10)...)
            @test nobs(train) == 135
            @test nobs(test)  == 15

            @test typeof(train) <: DataSubset
            @test typeof(test) <: DataSubset
        end
    end
end

println("<HEARTBEAT>")

@testset "kfolds" begin
    # FIXME: in 0.6 change back to (Xs,ys,vars...,tuples...)
    for var in (Xs, ys, Xv, yv, X, tuples[2])
        for kf in (@inferred(kfolds(var)),
                   @inferred(kfolds(var,ObsDim.Last())),
                   @inferred(kfolds(var,5)),
                   @inferred(kfolds(var,5,ObsDim.Last())),
                   kfolds(var,k=5,obsdim=:last))
            @test typeof(kf) <: FoldsView
            @test length(kf.train_indices) == length(kf.val_indices) == 5
            @test kf.data === var
            @test sum(length.(kf.val_indices)) == nobs(var)
        end
    end
    println("<HEARTBEAT>")
    for v1 in (X, Xv), v2 in (y, yv)
        kf = kfolds((v1, v2), k = 15)
        @test typeof(kf) <: FoldsView
        for ((train_x, train_y), (test_x, test_y)) in kf
            @test typeof(train_x) <: SubArray{Float64, 2}
            @test typeof(test_x) <: SubArray{Float64, 2}
            @test typeof(train_y) <: SubArray{String, 1}
            @test typeof(test_y) <: SubArray{String, 1}
            @test nobs(train_x) == 140
            @test nobs(train_y) == 140
            @test nobs(test_x) == 10
            @test nobs(test_y) == 10
        end
    end
    for var in (Xs, ys, vars...)
        kf = kfolds(var)
        @test typeof(kf) <: FoldsView
    end
end

println("<HEARTBEAT>")

@testset "leaveout" begin
    # FIXME: in 0.6 change back to (Xs,ys,vars...,tuples...)
    for var in (Xs, ys, Xv, yv, tuples[1])
        @test_throws ArgumentError leaveout(var,150)
        for kf in (@inferred(leaveout(var)),
                   @inferred(leaveout(var,ObsDim.Last())),
                   @inferred(leaveout(var,1)),
                   @inferred(leaveout(var,1,ObsDim.Last())),
                   leaveout(var,size=1,obsdim=:last))
            @test typeof(kf) <: FoldsView
            @test length(kf.train_indices) == length(kf.val_indices) == 150
            @test kf.data === var
            @test sum(length.(kf.val_indices)) == nobs(var)
        end
        for kf in (@inferred(leaveout(var,30)),
                   @inferred(leaveout(var,30,ObsDim.Last())),
                   leaveout(var,size=30,obsdim=:last))
            @test typeof(kf) <: FoldsView
            @test length(kf.train_indices) == length(kf.val_indices) == 5
            @test kf.data === var
            @test sum(length.(kf.val_indices)) == nobs(var)
        end
    end
    println("<HEARTBEAT>")
    for v1 in (X, Xv), v2 in (y, yv)
        @test length(leaveout((v1, v2)).val_indices) == 150
        kf = leaveout((v1, v2), size = 10)
        @test typeof(kf) <: FoldsView
        for ((train_x, train_y), (test_x, test_y)) in kf
            @test typeof(train_x) <: SubArray{Float64, 2}
            @test typeof(test_x) <: SubArray{Float64, 2}
            @test typeof(train_y) <: SubArray{String, 1}
            @test typeof(test_y) <: SubArray{String, 1}
            @test nobs(train_x) == 140
            @test nobs(train_y) == 140
            @test nobs(test_x) == 10
            @test nobs(test_y) == 10
        end
    end
    for var in (Xs, ys, vars...)
        kf = @inferred leaveout(var)
        @test typeof(kf) <: FoldsView
    end
end

println("<HEARTBEAT>")

@testset "nest DataView" begin
    # FIXME: in 0.6 change back to (Xs,ys,vars...,tuples...)
    for var in (Xs, ys, Xv, yv, tuples[2])
        A = ObsView(var)
        kf = @inferred kfolds(A,10)
        @test kf == kfolds(var,10)
        @test typeof(kf) == typeof(kfolds(var,10))

        A = BatchView(var,15)
        kf = @inferred kfolds(A,10)
        @test kf == kfolds(var,10)
        @test typeof(kf) == typeof(kfolds(var,10))

        A = BatchView(obsview(var),15)
        kf = @inferred kfolds(A,10)
        @test kf == kfolds(var,10)
        @test typeof(kf) == typeof(kfolds(var,10))
    end
end
